#amdclean

A build tool that converts AMD code to standard JavaScript.

[![Build Status](https://travis-ci.org/gfranko/amdclean.png?branch=master)](https://travis-ci.org/gfranko/amdclean)
[![NPM version](https://badge.fury.io/js/amdclean.png)](http://badge.fury.io/js/amdclean)

`npm install amdclean --save-dev`


## Use Case

**Single file** client-side JavaScript libraries or applications that want to use AMD to structure and build their code, but don't want an AMD footprint.


## Used By

* [Backbone-Require-Boilerplate](https://github.com/BoilerplateMVC/Backbone-Require-Boilerplate) - A Rad Backbone.js and Require.js Boilerplate Project

* [Ractive.js](http://www.ractivejs.org/) - Next-generation DOM manipulation

* [AddThis Smart Layers](https://www.addthis.com/get/smart-layers) - Third-party social widgets suite

* [Mod.js](http://madscript.com/modjs/) - JavaScript Workflow Tooling 


## Why

Many developers like to use the AMD API to write modular JavaScript, but do not want to include a full AMD loader (e.g. [require.js](https://github.com/jrburke/requirejs)), or AMD shim (e.g. [almond.js](https://github.com/jrburke/almond)) because of file size/source code readability.

By incorporating amdclean.js into the build process, there is no need for Require or Almond.

Since AMDclean rewrites your source code into standard JavaScript, it is a great
fit for JavaScript library authors who want a tiny download in one file after using the
[RequireJS Optimizer](http://requirejs.org/docs/optimization.html).

So, you get great code cleanliness with AMD, reduced file sizes, improved code readability, and easy integration with other developers who may not use AMD.


## Restrictions

**Note:** Same restrictions as almond.js.

It is best used for libraries or apps that use AMD and:

* optimize all the modules into one file -- no dynamic code loading.
* only have **one** require.config() call


##What is Supported

* `define()` and `require()` calls.

* [Shimmed modules](http://requirejs.org/docs/api.html#config-shim)

* [Simplified CJS wrapper](https://github.com/jrburke/requirejs/wiki/Differences-between-the-simplified-CommonJS-wrapper-and-standard-AMD-define#wiki-cjs)

* Exporting global modules to the global `window` object

* Storing all local modules inside of a single global object (Helps scoping issues for certain use cases)

## Download

Node - `npm install amdclean --save-dev`

Web - [Latest release](https://github.com/gfranko/amdclean/blob/master/src/amdclean.js)


## Usage

There are a few different ways that amdclean can be used including:

* With the RequireJS Optimizer (Grunt is also supported)

* As a standalone node module

* As a client-side library


###RequireJS Optimizer

* [Download the RequireJS optimizer](http://requirejs.org/docs/download.html#rjs).

* `npm install amdclean --save-dev`

* Make sure that each of your AMD modules have a module ID `path` alias name

```javascript
paths: {

	'first': '../modules/firstModule',

	'second': '../modules/secondModule',

	'third': '../modules/thirdModule'

}
```

* If you are **not** shimming any libraries, add an `onBuildWrite` config property to your RequireJS build configuration file.  Like this:

```javascript
onBuildWrite: function (moduleName, path, contents) {
    return module.require('amdclean').clean(contents);
}
```

* If you **are** shimming any libraries, add an `onModuleBundleComplete` config property to your RequireJS build configuration file instead.  Like this:

```javascript
onModuleBundleComplete: function (data) {
  var fs = require('fs'),
    amdclean = require('amdclean'),
    outputFile = data.path;
  fs.writeFileSync(outputFile, amdclean.clean({
    'code': fs.readFileSync(outputFile),
    'globalObject': true
  }));
}
```

* Run the optimizer using [Node](http://nodejs.org) (also [works in Java](https://github.com/jrburke/r.js/blob/master/README.md)).  More details can be found in the the [r.js](https://github.com/jrburke/r.js/) repo.

* If you are using the RequireJS optimizer [Grunt task](https://github.com/gruntjs/grunt-contrib-requirejs), then it is very easy to integrate amdclean using either the `onBuildWrite` or the `onModuleBundleComplete` config options. Here is an example Grunt file that includes the RequireJS optimizer plugin with amdclean support:

```javascript
module.exports = function(grunt) {
  grunt.initConfig({
    pkg: grunt.file.readJSON('package.json'),
    requirejs: {
      js: {
        options: {
          findNestedDependencies: true,
          baseUrl: 'src/js/app/modules',
          wrap: true,
          preserveLicenseComments: false,
          optimize: 'none',
          mainConfigFile: 'src/js/app/config/config.js',
          include: ['first'],
          out: 'src/js/app/exampleLib.js',
          onBuildWrite: function( name, path, contents ) {
            return require('amdclean').clean(contents);
          }
        }
      }
    }
  });
  grunt.loadNpmTasks('grunt-contrib-requirejs');
  grunt.registerTask('build', ['requirejs:js']);
  grunt.registerTask('default', ['build']);
};
```

###Node Module

* `npm install amdclean --save-dev`

* Require the module

```javascript
var cleanAMD = require('amdclean');
```

* Call the clean method

```javascript
var code = 'define("exampleModule", function() {});'
var cleanedCode = cleanAMD.clean(code);
```


###Client-side Library

* Include all dependencies

```html
<script src="http://esprima.org/esprima.js"></script>
<script src="http://esprima.org/test/3rdparty/escodegen.browser.js"></script>
<script src="http://gregfranko.com/javascripts/estraverse.js"></script>
<script src="http://cdnjs.cloudflare.com/ajax/libs/lodash.js/2.2.1/lodash.js"></script>
<script src="http://gregfranko.com/javascripts/amdclean.js"></script>
```

* Use the global `amdclean` object and `clean()` method

```javascript
var cleanedCode = amdclean.clean('define("example", [], function() { var a = true; });');
```

## Requirements

* [Esprima](https://github.com/ariya/esprima) 1.0+

* [Lodash](https://github.com/lodash/lodash) 2.2.1+

* [Estraverse](https://github.com/Constellation/estraverse) 1.3.1+

* [Escodegen](https://github.com/Constellation/escodegen) 0.0.27+

## Optional Dependencies

* [r.js](https://github.com/jrburke/r.js/) 2.1.0+


## How it works

amdclean uses Esprima to generate an AST (Abstract Syntax Tree) from the provided source code, estraverse to traverse and update the AST, and escodegen to generate the new standard JavaScript code.  There are a few different techniques that amdclean uses to convert AMD to standard JavaScript code:


###Define Calls

_AMD_

```javascript
define('example', [], function() {

});
```

_Standard_

```javascript
var example = function () {

}();
```

---

_AMD_

```javascript
define('example', [], function() {
  return function(name) {
    return 'Hello ' + name;
  };
});
```

_Standard_

```javascript
var example = function (name) {
  return 'Hello ' + name;
};
```

---

_AMD_

```javascript
define('example', [], function() {
  return 'I love AMDClean';
});
```

_Standard_

```javascript
var example = 'I love AMDClean';
```

---

_AMD_

```javascript
define('example', ['example1', 'example2'], function(one, two) {

});
```


_Standard_

```javascript
var example = function (one, two) {

}(example1, example2);
```

---

_AMD_

```javascript
define("backbone", ["underscore","jquery"], (function (global) {
    return function () {
        var ret, fn;
        return ret || global.Backbone;
    };
}(this)));
```


_Standard_

```javascript
var backbone = window.Backbone;
```

---

_AMD_

```javascript
define('third',{
	exampleProp: 'This is an example'
});
```

_Standard_

```javascript
var third = {
	exampleProp: 'This is an example'
};
```

---

###Require Calls

**Note:** `require(['someModule'])` calls are removed from the built source code

_AMD_

```javascript
require([], function() {
	var example = true;
});
```

_Standard_

```javascript
(function () {
    var example = true;
}());
```

---

_AMD_

```javascript
require(['anotherModule'], function(anotherModule) {
	var example = true;
});
```

_Standard_

```javascript
(function (anotherModule) {
    var example = true;
}(anotherModule));
```


##Options

The amdclean `clean()` method accepts a string or an object.  Below is an example object with all of the available configuration options:

```javascript
amdclean.clean({
  // The source code you would like to be 'cleaned'
  code: 'define("example", [], function(one, two) {});',
  // The relative file path of the file to be cleaned.  Use this option if you are not using the code option.
  // Hint: Use the __dirname trick
  filePath: '',
  // The modules that you would like to set as window properties
  // An array of strings (module names)
  globalModules: [],
  // Determines if all of your local modules are stored in a single global object (helps with scoping in certain cases)
  globalObject: false,
  // Determines the name of your global object that stores all of your global modules
  globalObjectName: 'amdclean',
  // All esprima API options are supported: http://esprima.org/doc/
  esprima: {},
  // All escodegen API options are supported: https://github.com/Constellation/escodegen/wiki/API
  escodegen: {},
  // If there is a comment (that contains the following text) on the same line or one line above a specific module, the module will not be removed
  commentCleanName: 'amdclean',
  // The ids of all of the modules that you would not like to be removed
  ignoreModules: [],
  // Determines if all of the require() method calls will be removed
  removeAllRequires: false,
  // Determines if all of the 'use strict' statements will be removed
  removeUseStricts: true,
  // Allows you to pass an expression that will override shimmed modules return values
  // e.g. { 'backbone': 'window.Backbone' }
  shimOverrides: {},
  // Prevents multiple global objects from being instantiated when using the onBuildWrite Require.js hook
  // Set this to false if you are using AMDClean for more than one build AND
  // are using the onModuleBundleComplete Require.js hook
  rememberGlobalObject: true
})
```


## Unit Tests

All unit tests are written using the [jasmine-node](https://github.com/mhevery/jasmine-node) library and can be found in the `test/specs/` folder.  You can run the unit tests by typing: `npm test`.

## Contributing

Please send all PR's to the `dev` branch.

If your PR is a code change:

1.  Update `amdclean.js` inside of the `src` directory.
2.  Add a Jasmine unit test to `convert.js` inside of the `test/specs` folder
3.  Install all node.js dev dependencies: `npm install`
4.  Install gulp.js globally: `sudo npm install gulp -g`
5.  Lint, Minify, and Run all unit tests with Gulp: `gulp`
6.  Verify that the minified output file has been updated in `build/amdclean.min.js`
7.  Send the PR!

**Note:** There is a gulp `watch` set up called, `amdclean-watch`, that will automatically lint, minify, and run all the AMDClean unit tests when `src/amdclean.js` is changed.  Feel free to use it.


## FAQ

__Why would I use amdclean instead of Almond.js?__

 - Although Almond is very small (~1k gzipped and minified), most JavaScript library authors do not want to have to include it in their library's source code.  If you are not using an AMD plugin, then amdclean provides the benefit of AMD without increasing your library's file size.

__AMDClean does not seem to be cleaning shimmed modules.  What am I doing wrong?__

 - Since Require.js does not expose the [shim](http://requirejs.org/docs/api.html#config-shim) functionality within the `onBuildWrite` config property, you must use the `onModuleBundleComplete` config property instead.  Like this:

 ```javascript
onModuleBundleComplete: function (data) {
  var fs = require('fs'),
    amdclean = require('amdclean'),
    outputFile = data.path;
  fs.writeFileSync(outputFile, amdclean.clean({
    'code': fs.readFileSync(outputFile),
    'globalObject': true
  }));
}
 ```

__What if I don't want all define() and require() method calls to be removed?__

 - If you don't want one or more define() and require() methods to be removed by `amdclean`, you have a few options.  If the module has a named module id associated with it, then you can add the associated module id to the `ignoreModules` option array.  Like this:

 ```javascript
var amdclean = require('amdclean');
amdclean.clean({
    'code': 'define("randomExample", function() { console.log("I am a random example"); });',
    'ignoreModules': ['randomExample']
});
 ```

 If there is not an associated module id, then you must put a comment with only the words _amdclean_ on the same line or one line above the method in question.  For example, `amdclean` would not remove the `define()` method below:

 ```javascript
// amdclean
define('example', [], function() {});
 ```

If you want to use different text than `amdclean`, you can customize the comment name by using the `commentCleanName` option.

__Why are define() method placeholder functions inserted into my source?__

- This is the default behavior of r.js when a module(s) is not wrapped in a define() method.  Luckily,  this behavior can be overridden by setting the `skipModuleInsertion` option to `true` in your build configuration.

__How would I expose one or more modules as a global window property?__

- You can use the `globalModules` option to list all of the modules that you would like to expose as a `window` property

__I am having a scope problem with all of the local module variables.  What can I do?__

- You can use the `globalObject` option to store all of your modules in a single global object that uses the top-most function scope.  You can even name that global object whatever you prefer by using the `globalObjectName` option.


## License

Copyright (c) 2014 Greg Franko Licensed under the MIT license.
